from pathlib import Path

import pkg_resources

CONFIG_INI_FILENAME = 'config.ini'
STYLES_TEMPLATE_FILENAME = 'styles.template'
DEFAULT_PATH_TO_CONFIG = Path(pkg_resources.resource_filename('pyreball', 'cfg'))
PATH_TO_CONFIG_LOCATION = Path(pkg_resources.resource_filename('pyreball', 'config_location'))



import configparser
import logging
import os
from pathlib import Path

import pytest
from unittest import mock

from pyreball.utils.utils import (check_choice_string_parameter, check_integer_within_range, read_file_config,
                                  get_file_config, check_and_fix_parameters, merge_parameter_dictionaries,
                                  merge_values, ChoiceParameter, IntegerParameter, make_sure_dir_exists, _map_env_value,
                                  get_parameter_value, carefully_remove_directory_if_exists, Substitutor)


@pytest.fixture
def simple_parameter_specifications():
    return [
        ChoiceParameter('--align', choices=['left', 'center', 'right'], default='center', help=''),
        ChoiceParameter('--do-stuff', choices=['yes', 'no'], default='yes', help=''),
        ChoiceParameter('--highlight', choices=['yes', 'no'], default='no', help=''),
        ChoiceParameter('--organize', choices=['yes', 'no'], default='no', help=''),
        IntegerParameter('--page-size', boundaries=(10, 20), default=15, help=''),
    ]


@pytest.mark.parametrize('value,none_allowed,warn_msg,err_msg,expected_result', [
    ('one', True, None, None, 'one'),
    ('one', False, None, None, 'one'),
    (None, True, None, None, None),
    (None, False, 'Parameter param1 was not set', None, 'three'),
    ('four', True, None, 'Parameter param1 is set to an unsupported', 'four'),
    ('four', False, None, 'Parameter param1 is set to an unsupported', 'four'),
])
def test_check_choice_string_parameter(value, none_allowed, warn_msg, err_msg, expected_result):
    warning_messages = []
    error_messages = []
    result_value = check_choice_string_parameter('param1', value, ['one', 'two'], 'three',
                                                 none_allowed, warning_messages, error_messages)
    assert result_value == expected_result
    contains_warning_msg = any(map(lambda msg: warn_msg in msg, warning_messages))
    contains_error_msg = any(map(lambda msg: err_msg in msg, error_messages))
    assert contains_warning_msg if warn_msg else not contains_warning_msg
    assert contains_error_msg if err_msg else not contains_error_msg


@pytest.mark.parametrize('value,low,high,none_allowed,warn_msg,err_msg,expected_result', [
    (25, 20, 30, True, None, None, 25),
    (25, 20, 30, False, None, None, 25),
    (None, 20, 30, True, None, None, None),
    (None, 20, 30, False, None, 'Could not parse param1', None),
    ('four', 20, 30, True, None, 'Could not parse param1', 'four'),
    ('four', 20, 30, False, None, 'Could not parse param1', 'four'),
    (10, 20, 30, True, 'Parameter param1 is less than 20, setting it to 20.', None, 20),
    (50, 20, 30, False, 'Parameter param1 is more than 30, setting it to 30.', None, 30),
    (25, None, None, False, None, None, 25),
    (25, None, 20, False, 'Parameter param1 is more than 20, setting it to 20.', None, 20),
    (25, 30, None, False, 'Parameter param1 is less than 30, setting it to 30.', None, 30),
])
def test_check_integer_within_range(value, low, high, none_allowed, warn_msg, err_msg, expected_result):
    warning_messages = []
    error_messages = []
    result_value = check_integer_within_range('param1', value, low, high, none_allowed,
                                              warning_messages, error_messages)
    assert result_value == expected_result
    contains_warning_msg = any(map(lambda msg: warn_msg in msg, warning_messages))
    contains_error_msg = any(map(lambda msg: err_msg in msg, error_messages))
    assert contains_warning_msg if warn_msg else not contains_warning_msg
    assert contains_error_msg if err_msg else not contains_error_msg


@pytest.mark.parametrize("parameters,none_allowed,warn_msgs,err_msgs,expected_result", [
    # everything is fine
    (
            {
                'align': 'center',
                'do_stuff': 'no',
                'highlight': 'yes',
                'organize': 'yes',
                'page_size': 10,
            },
            True, [], [],
            {
                'align': 'center',
                'do_stuff': 'no',
                'highlight': 'yes',
                'organize': 'yes',
                'page_size': 10,
            }
    ),
    # None values are used for the missing ones
    (

            {
                'do_stuff': 'no',
                'highlight': 'yes',
                'page_size': 30,
            },
            True, ['Parameter page_size is more than'], [],
            {
                'align': None,
                'do_stuff': 'no',
                'highlight': 'yes',
                'organize': None,
                'page_size': 20,
            }
    ),
    # Default values are used for the missing ones
    (
            {
                'do_stuff': 'no',
                'highlight': 'yes',
                'page_size': 4,
            },
            False, ['Parameter align was not set', 'Parameter organize was not set',
                    'Parameter page_size is less than'], [],
            {
                'align': 'center',
                'do_stuff': 'no',
                'highlight': 'yes',
                'organize': 'no',
                'page_size': 10,
            }
    ),
    # Unsupported values
    (
            {
                'do_stuff': 'maybe',
                'highlight': 'yes',
                'organize': 'no',
                'page_size': 'not_a_number',
            },
            False, ['Parameter align was not set'], ['Parameter do_stuff is set to an unsupported',
                                                     'Could not parse page_size parameter as an integer'],
            "result_does_not_matter"
    ),
])
def test_check_and_fix_parameters(parameters, none_allowed, warn_msgs, err_msgs, expected_result, caplog,
                                  simple_parameter_specifications):
    caplog.set_level(logging.WARNING)
    if len(err_msgs) > 0:
        with pytest.raises(SystemExit):
            check_and_fix_parameters(parameters=parameters, parameter_specifications=simple_parameter_specifications,
                                     none_allowed=none_allowed)
        for msg in warn_msgs:
            assert msg in caplog.text
        for msg in err_msgs:
            assert msg in caplog.text
    else:
        result_parameters = check_and_fix_parameters(parameters=parameters,
                                                     parameter_specifications=simple_parameter_specifications,
                                                     none_allowed=none_allowed)
        for msg in warn_msgs:
            assert msg in caplog.text
        assert result_parameters == expected_result


def test_read_file_config__existing_file(tmpdir):
    filename = "conf"
    directory = Path(tmpdir)
    with open(directory / filename, 'w') as f:
        f.writelines(['[Parameters]\n', 'a = b\n'])
    config = read_file_config(filename, directory)
    assert config['Parameters']['a'] == 'b'


def test_read_file_config__non_existing_file(tmpdir, caplog):
    caplog.set_level(logging.ERROR)
    filename = "conf"
    directory = Path(tmpdir)
    with pytest.raises(SystemExit):
        read_file_config(filename, directory)
    assert 'Could not find file' in caplog.text


def test_get_file_config__correct_specification(simple_parameter_specifications):
    config = configparser.ConfigParser()
    # 'align' is missing and 'page-size' is incorrect
    config['Parameters'] = {
        'do-stuff': 'no',
        'highlight': 'yes',
        'organize': 'yes',
        'page-size': 22,
    }
    expected_config_parameters = {
        'align': 'center',
        'do_stuff': 'no',
        'highlight': 'yes',
        'organize': 'yes',
        'page_size': 20,
    }
    with mock.patch('pyreball.utils.utils.read_file_config', return_value=config):
        config_parameters = get_file_config(filename="arbitrary",
                                            parameter_specifications=simple_parameter_specifications,
                                            directory=Path("/arbitrary"))
        assert config_parameters == expected_config_parameters


def test_get_file_config__incorrect_specification(caplog, simple_parameter_specifications):
    caplog.set_level(logging.ERROR)
    config = configparser.ConfigParser()
    config['unknown'] = {'organize': 'yes'}
    with mock.patch('pyreball.utils.utils.read_file_config', return_value=config):
        with pytest.raises(SystemExit):
            get_file_config(filename="arbitrary", parameter_specifications=simple_parameter_specifications,
                            directory=Path("/arbitrary"))
        assert 'Parameters section not found in' in caplog.text


@pytest.mark.parametrize("test_input_1,test_input_2,expected_result", [
    (None, None, None),
    ('a', None, 'a'),
    (None, 'b', 'b'),
    ('a', 'b', 'a'),
])
def test_merge_values(test_input_1, test_input_2, expected_result):
    assert merge_values(test_input_1, test_input_2) == expected_result


@pytest.mark.parametrize("primary_parameters,secondary_parameters,expected_result", [
    (
            {
                'align': None,
                'do_stuff': 'no',
                'highlight': 'yes',
                'organize': None,
                'page_size': 10,
            },
            {
                'align': 'center',
                'do_stuff': None,
                'highlight': 'no',
                'organize': None,
                'page_size': 20,
            },
            {
                'align': 'center',
                'do_stuff': 'no',
                'highlight': 'yes',
                'organize': None,
                'page_size': 10,
            }
    ),
])
def test_merge_parameter_dictionaries(primary_parameters, secondary_parameters, expected_result,
                                      simple_parameter_specifications):
    assert merge_parameter_dictionaries(primary_parameters, secondary_parameters,
                                        parameter_specifications=simple_parameter_specifications) == expected_result


@pytest.mark.parametrize("test_input,expected_result", [
    ('None', None),
    ('yes', True),
    ('no', False),
    ('anything', 'anything'),
    (42, 42),
])
def test_map_env_value(test_input, expected_result):
    assert _map_env_value(test_input) == expected_result


def test_get_parameter_value():
    pars = '{"a": 2, "html_dir_path": "/tmp/dir"}'
    with mock.patch.dict(os.environ, {'_TMP_PYREBALL_GENERATOR_PARAMETERS': pars}):
        assert get_parameter_value("a") == 2
        assert get_parameter_value("html_dir_path") == "/tmp/dir"
        assert get_parameter_value("html_dir_name") == "dir"
        assert get_parameter_value("html_file_path") == "/tmp/dir.html"


def test_make_sure_dir_exists(tmpdir):
    directory = str(tmpdir / 'mydir')
    assert not os.path.exists(directory)
    make_sure_dir_exists(directory)
    assert os.path.exists(directory)
    # when executed again, nothing breaks
    make_sure_dir_exists(directory)
    assert os.path.exists(directory)


@mock.patch('pyreball.utils.utils.shutil')
def test_carefully_remove_directory_if_exists__doesnt_exist(shutil_mock, tmpdir):
    directory = Path(tmpdir / 'whatever')
    assert not directory.exists()
    carefully_remove_directory_if_exists(directory)
    shutil_mock.rmtree.assert_not_called()


@mock.patch('pyreball.utils.utils.shutil')
def test_carefully_remove_directory_if_exists__error_when_deleting(shutil_mock, tmpdir):
    shutil_mock.rmtree.side_effect = OSError
    directory = Path(tmpdir / 'whatever')
    directory.mkdir(parents=True)
    assert directory.exists()
    with pytest.raises(OSError):
        carefully_remove_directory_if_exists(directory)


def test_carefully_remove_directory_if_exists__empty_dir(tmpdir):
    directory = Path(tmpdir / 'whatever')
    directory.mkdir(parents=True)
    assert directory.exists()
    carefully_remove_directory_if_exists(directory)
    assert not directory.exists()


def test_carefully_remove_directory_if_exists__with_image_files(tmpdir):
    directory = Path(tmpdir / 'whatever')
    directory.mkdir(parents=True)

    # Creates empty files
    filenames = ['img.png', 'img.jpg', 'img.svg']
    for filename in filenames:
        with open(directory / filename, 'w') as fp:
            pass

    assert directory.exists()
    assert len(list(directory.glob("*"))) == len(filenames)
    carefully_remove_directory_if_exists(directory)
    assert not directory.exists()


def test_carefully_remove_directory_if_exists__with_non_image_files(tmpdir):
    directory = Path(tmpdir / 'whatever')
    directory.mkdir(parents=True)

    # Creates empty files
    filenames = ['img.png', 'img.jpg', 'img.svg', 'important_script.py']
    for filename in filenames:
        with open(directory / filename, 'w') as fp:
            pass

    assert directory.exists()
    assert len(list(directory.glob("*"))) == len(filenames)
    with pytest.raises(ValueError) as excinfo:
        carefully_remove_directory_if_exists(directory)
    assert "Cannot delete the original html directory" in str(excinfo.value)
    assert directory.exists()


@pytest.mark.parametrize("text,repls,expected_result", [
    (
        '',
        [('apple', 'pear'), ('dark', 'light'), ('red', 'green')],
        ''
    ),
    (
        'This apple is dark red',
        [('apple', 'pear'), ('dark', 'light'), ('red', 'green')],
        'This pear is light green'
    ),
    (
        'This apple is dark red',
        [('apple', 'pear'), ('dark', 'light'), ('red', 'green'), ('pear', 'orange')],
        'This orange is light green'
    ),
    (
        'my aa and also aab ccdd cc table-5035059472 and also table-5035059472-6',
        [(r'aab?', 'XX'), (r'cc(dd)?', 'YY'), ('table-5035059472(-6)?', 'table-6')],
        'my XX and also XX YY YY table-6 and also table-6',
     )
])
def test_substitutor(text, repls, expected_result):
    substitutor = Substitutor(repls)
    result = substitutor.sub(text)
    assert result == expected_result



"""Text utils that can be used to prepare strings from formatting, list and link tags."""
from typing import Any
from pyreball.utils.utils import get_parameter_value


def bold(value: Any) -> str:
    if get_parameter_value('html_file_path'):
        return "<b>" + str(value) + "</b>"
    else:
        return str(value)


def em(value: Any) -> str:
    if get_parameter_value('html_file_path'):
        return "<em>" + str(value) + "</em>"
    else:
        return str(value)


def code(value: Any) -> str:
    if get_parameter_value('html_file_path'):
        return "<code>" + str(value) + "</code>"
    else:
        return str(value)


def ul(*args: Any) -> str:
    if get_parameter_value('html_file_path'):
        result = "<ul>"
        for value in args:
            if isinstance(value, str) and (value.startswith("<ol>") or value.startswith("<ul>")):
                # don't append <li> if it is a nested list
                result += str(value)
            else:
                result += "<li>" + str(value) + "</li>"
        result += "</ul>"
        return result
    else:
        return str(list(args))


def ol(*args: Any) -> str:
    if get_parameter_value('html_file_path'):
        result = "<ol>"
        for value in args:
            if isinstance(value, str) and (value.startswith("<ol>") or value.startswith("<ul>")):
                # don't append <li> if it is a nested list
                result += str(value)
            else:
                result += "<li>" + str(value) + "</li>"
        result += "</ol>"
        return result
    else:
        return str(list(args))


def link(text: str, url: str) -> str:
    return f'<a href="{url}">{text}</a>'



__version__ = "0.1.0"

from pyreball.html import (set_title, print_h1, print_h2, print_h3, print_h4, print_h5, print_h6,
                           print_div, print_code, print_html, print_table, plot_graph, plot_multi_graph,
                           create_reference)
from pyreball.text import bold, em, code, ul, ol, link



"""Main functions that serve as building blocks of the final html file."""
import os
import re
import random
import io
from typing import Optional, List, Tuple, Any, Union, cast
from typing import TYPE_CHECKING

from pyreball.utils.utils import merge_values, make_sure_dir_exists, get_parameter_value

if TYPE_CHECKING:
    # needed for mypy
    # noinspection PyUnresolvedReferences,PyPackageRequirements
    import altair  # type: ignore
    # noinspection PyPackageRequirements
    import matplotlib  # type: ignore
    # noinspection PyPackageRequirements
    import pandas  # type: ignore
    # noinspection PyPackageRequirements
    import plotly  # type: ignore
    # noinspection PyPackageRequirements
    import bokeh  # type: ignore

AltairFigType = Union['altair.vegalite.v4.api.Chart',
                      'altair.vegalite.v4.api.ConcatChart',
                      'altair.vegalite.v4.api.FacetChart',
                      'altair.vegalite.v4.api.HConcatChart',
                      'altair.vegalite.v4.api.LayerChart',
                      'altair.vegalite.v4.api.RepeatChart',
                      'altair.vegalite.v4.api.VConcatChart']
FigType = Union['matplotlib.figure.Figure',
                'plotly.graph_objs._figure.Figure',
                'altair.vegalite.v4.api.Chart',
                'altair.vegalite.v4.api.ConcatChart',
                'altair.vegalite.v4.api.FacetChart',
                'altair.vegalite.v4.api.HConcatChart',
                'altair.vegalite.v4.api.LayerChart',
                'altair.vegalite.v4.api.RepeatChart',
                'altair.vegalite.v4.api.VConcatChart']


class Reference:

    def __init__(self, default_text: Optional[str] = None) -> None:
        self.id = "id" + str(random.getrandbits(64))
        self.text = default_text

    def __str__(self) -> str:
        return '<a href="#ref-' + self.id + '">' + (self.id if self.text is None else self.text) + '</a>'

    def __call__(self, text: str):
        return '<a href="#ref-' + self.id + '">' + text + '</a>'


def create_reference(default_text: Optional[str] = None) -> Reference:
    return Reference(default_text)


def _check_and_mark_reference(reference: Reference) -> None:
    """Check and save a reference.

    This function is used when references are added to tables or plots.
    If a table or plot is about to get a reference that was already used for another object, an error is raised.
    """
    if not hasattr(_check_and_mark_reference, 'used_references'):
        _check_and_mark_reference.used_references = set()

    if reference.id in _check_and_mark_reference.used_references:
        raise ValueError("Reference is used for the second time. You have to create another reference for this object.")
    else:
        _check_and_mark_reference.used_references.add(reference.id)


def set_title(title: str) -> None:
    """Set page title.

    Note that this function does not have to be called at the beginning of the script.
    If this function is not called via pyreball and parameter keep_stdout is set to True,
    it just prints the title to stdout.

    :param title: title string.
    """
    if not get_parameter_value('html_file_path') or get_parameter_value('keep_stdout'):
        print(title)
    if get_parameter_value('html_file_path'):
        # it is assumed that the heading is already written into the file,
        # so find the line with title element and replace its contents
        with open(get_parameter_value('html_file_path'), "r") as f:
            lines = f.readlines()

        # replace the title and also add "custom" class so that we know it was replaced by this function
        lines = [re.sub(r'^<title>[^<]*</title>', f'<title class="custom">{title}</title>', line) for line in lines]
        with open(get_parameter_value('html_file_path'), "w") as f:
            f.writelines(lines)


def _write_to_html(string: str) -> None:
    if get_parameter_value('html_file_path'):
        with open(get_parameter_value('html_file_path'), 'a') as f:
            f.write(string)
            f.write("\n")


def _tidy_title(title: str) -> str:
    """
    Transforms title string into lowercase alphanumerical sequence separated by underscores.

    :param title: str
    :return: str
    """
    new_title = re.sub("([a-z])([A-Z])", r"\g<1> \g<2>", title)  # CamelCase -> words with spaces
    new_title = new_title.lower()  # to lower case
    new_title = re.sub(r'%', '_percent', new_title)  # replace % symbol with "_percent"
    new_title = re.sub(r'[=-\\.,]', ' ', new_title)  # replace some special characters with spaces
    new_title = re.sub(r'[^\w\s]', '', new_title).strip()  # remove remaining special characters
    new_title = re.sub(r'\s+', '_', new_title)  # replace sequences of spaces with an underscore
    return new_title


def _reduce_whitespaces(string: str) -> str:
    string = re.sub(r'\s+', ' ', string)
    return string.strip()


def _get_heading_number(level: int, l_heading_counting: List[int]) -> str:
    return ".".join(map(str, l_heading_counting[:level]))


def _print_heading(string: str, level: int = 1) -> None:
    if level > 6:
        raise ValueError("Heading level cannot be greater than 6.")
    if level < 1:
        raise ValueError("Heading level cannot be less than 1.")

    if not get_parameter_value('html_file_path') or get_parameter_value('keep_stdout'):
        print("#" * level + " " + str(string))

    if get_parameter_value('html_file_path'):
        if not hasattr(_print_heading, 'heading_index'):
            _print_heading.heading_index = 1
        heading_index = _print_heading.heading_index

        if get_parameter_value('numbered_headings'):
            if not hasattr(_print_heading, 'heading_counting'):
                # what is the index of current h1, h2, h3, h4, h5, h6?
                _print_heading.heading_counting = [0, 0, 0, 0, 0, 0]

            # increase the number in the level
            _print_heading.heading_counting[level - 1] = _print_heading.heading_counting[level - 1] + 1
            # reset all sub-levels
            _print_heading.heading_counting[level:] = [0] * (6 - level)
            # get the string of the numbered section and append non-breakable space
            non_breakable_spaces = '\u00A0\u00A0'
            heading_number_str = _get_heading_number(level, _print_heading.heading_counting) + non_breakable_spaces
        else:
            heading_number_str = ''

        string = heading_number_str + _reduce_whitespaces(string)
        # use heading_index in the id of the heading so there are no collisions in the case of same texts
        tidy_string = _tidy_title(string) + "_" + str(heading_index)
        pilcrow_sign = "\u00B6"
        string = string + f'<a class="anchor-link" href="#{tidy_string}">{pilcrow_sign}</a>'
        _write_to_html(f'<h{level} id="{tidy_string}">{string}</h{level}>')
        _print_heading.heading_index += 1


def print_h1(string: str) -> None:
    _print_heading(string, level=1)


def print_h2(string: str) -> None:
    _print_heading(string, level=2)


def print_h3(string: str) -> None:
    _print_heading(string, level=3)


def print_h4(string: str) -> None:
    _print_heading(string, level=4)


def print_h5(string: str) -> None:
    _print_heading(string, level=5)


def print_h6(string: str) -> None:
    _print_heading(string, level=6)


def print_div(*string: Any, sep: str = "") -> None:
    if not get_parameter_value('html_file_path') or get_parameter_value('keep_stdout'):
        print(*string, sep=sep)
    if get_parameter_value('html_file_path'):
        html_div = re.sub(r'\n', '<br>', sep.join(map(str, string)))
        _write_to_html(f"<div>{html_div}</div>")


def print_code(string: str, highlight_syntax: bool = True) -> None:
    if not get_parameter_value('html_file_path') or get_parameter_value('keep_stdout'):
        print(string)
    if get_parameter_value('html_file_path'):
        if highlight_syntax:
            _write_to_html('<pre class="prettyprint lang-py">' + string + '</pre>')
        else:
            _write_to_html('<pre>' + string + '</pre>')


def print_html(string: str) -> None:
    if not get_parameter_value('html_file_path') or get_parameter_value('keep_stdout'):
        print(string)
    if get_parameter_value('html_file_path'):
        _write_to_html(string)


def _prepare_caption_element(prefix: str, caption: Optional[str], numbered: bool, index: int, anchor_link: str) -> str:
    if numbered:
        caption_text = f'{prefix} {index}'
        if caption:
            caption_text = f'{caption_text}: {caption}'
        else:
            caption_text = f'{caption_text}.'
    elif caption:
        caption_text = f'{caption}'
    else:
        caption_text = ''
    return f'\n<div class="text-centered"><a name="{anchor_link}"><b>\n{caption_text}\n</b></a></div>\n'


def _prepare_table_html(df: 'pandas.DataFrame',
                        caption: Optional[str] = None,
                        align: str = "center",
                        full_table: bool = True,
                        numbered: bool = True,
                        reference: Optional[Reference] = None,
                        sortable: bool = False,
                        tab_index: int = 0,
                        sorting_definition: Optional[Tuple[str, str]] = None,
                        **kwargs: Any) -> str:
    align_mapping = {'center': 'centered', 'left': 'left-aligned', 'right': 'right-aligned'}

    if sorting_definition:
        # if we have sorting definition, we turn on sortable
        sortable = True
    table_classes = ["centered"]
    if sortable and not sorting_definition:
        # add this class only to sortable tables that don't have sorting definition
        table_classes.append("sortable_table")
    table_html = df.to_html(classes=table_classes, **kwargs)
    if reference:
        _check_and_mark_reference(reference)
        anchor_link = 'table-' + reference.id + '-' + str(tab_index)
    else:
        anchor_link = 'table-' + str(tab_index)

    caption_element = _prepare_caption_element(prefix='Table', caption=caption, numbered=numbered,
                                               index=tab_index, anchor_link=anchor_link)
    # div that is scrollable
    scroller_id = 'table-scroller-' + str(tab_index)
    # div button that expands the table
    expander_id = 'table-expander-' + str(tab_index)
    if full_table:
        table_html = (caption_element + '<div id="' + scroller_id + '" class="table-scroller">\n'
                      + table_html + '\n</div>\n')
        table_html += ('<div class="text-centered table-expander" style="display: none;" id="' + expander_id
                       + '" onclick="change_expand(this, \'' + scroller_id + '\')">⟱</div>')
    else:
        table_html = (caption_element + '<div id="' + scroller_id + '" class="table-scroller-collapsed">\n'
                      + table_html + '\n</div>\n')
        table_html += '<div class="text-centered table-expander" ' \
                      'id="' + expander_id + '" onclick="change_expand(this, \'' + scroller_id + '\')">⟱</div>'

    table_html = f'<div class="table-wrapper-inner {align_mapping[align]}">' + table_html + '</div>'
    table_html = '<div class="table-wrapper">' + table_html + '</div>'
    if sorting_definition:
        if sorting_definition[0] not in df.columns:
            raise ValueError(f"{sorting_definition[0]} is not a column in provided data frame.")
        if sorting_definition[1] not in ['asc', 'desc']:
            raise ValueError(f"sorting_definition must be either None or a pair (<column_name>, <sorting>), "
                             f"where <sorting> is either 'asc' or 'desc'.")
        column_index = df.columns.get_loc(sorting_definition[0]) + 1  # (+ index)
        table_init = '{"retrieve": true, "paging": false, "searching": false, "info": false}'
        js = f"var table = $('#{scroller_id} > table').DataTable({table_init});" \
             f"table.order( [ {column_index}, '{sorting_definition[1]}' ] ).draw();"
        table_html += f"\n<script>{js}</script>"
    return table_html


def print_table(df: 'pandas.DataFrame',
                caption: Optional[str] = None,
                reference: Optional[Reference] = None,
                align: Optional[str] = None,
                numbered: Optional[bool] = None,
                full_table: Optional[bool] = None,
                sortable: Optional[bool] = None,
                sorting_definition: Optional[Tuple[str, str]] = None,
                **kwargs: Any) -> None:
    """Print pandas DataFrame into HTML.

    The sortable tables are based on https://datatables.net/examples/basic_init/zero_configuration.html.

    :param df: data frame to be printed.
    :param caption: text caption.
    :param reference: reference object.
    :param align: how to align the table: "left", "center", "right".
    If None, settings from config or CLI arguments is used.
    :param numbered: should be the caption be numbered?
    :param full_table: whether to show the table expanded. If None, settings from config or CLI arguments is used.
    :param sortable: whether to allow sortable columns. If None, settings from config or CLI arguments is used.
    :param sorting_definition: how to sort the table initially, in the form (<column_name>, <sorting>),
    where <sorting> is either 'asc' or 'desc'.
    :param kwargs: other parameters to pandas to_html method.
    :return:
    """
    if not get_parameter_value('html_file_path') or get_parameter_value('keep_stdout'):
        print(df)
    if get_parameter_value('html_file_path'):

        if not hasattr(print_table, 'table_index'):
            print_table.table_index = 1
        table_index = print_table.table_index

        align = cast(str, merge_values(primary_value=align, secondary_value=get_parameter_value('align_tables')))
        numbered = bool(merge_values(primary_value=numbered, secondary_value=get_parameter_value('numbered_tables')))
        sortable = bool(merge_values(primary_value=sortable, secondary_value=get_parameter_value('sortable_tables')))
        full_table = bool(merge_values(primary_value=full_table, secondary_value=get_parameter_value('full_tables')))
        table_html = _prepare_table_html(df=df, caption=caption, align=align, full_table=full_table,
                                         numbered=numbered, reference=reference, sortable=sortable,
                                         tab_index=table_index, sorting_definition=sorting_definition, **kwargs)
        _write_to_html(table_html)
        print_table.table_index += 1


def _construct_plot_anchor_link(reference: Optional[Reference], plot_ind: int) -> str:
    if reference:
        _check_and_mark_reference(reference)
        return 'img-' + reference.id + '-' + str(plot_ind)
    else:
        return 'img-' + str(plot_ind)


def _wrap_plot_element_by_outer_divs(img_element: str, align: str, hidden: bool) -> str:
    img_element = f'<div align="{align}"><div style="display: inline-block;">' + img_element + '</div></div>'
    if hidden:
        return f'<div class="image-wrapper" style="display: none;">' + img_element + '</div>'
    else:
        return f'<div class="image-wrapper">' + img_element + '</div>'


def _prepare_matplotlib_plot_element(fig: 'matplotlib.figure.Figure',
                                     l_plot_index: int,
                                     plot_format: Optional[str] = None,
                                     embedded: Optional[bool] = None) -> str:
    if plot_format is not None and plot_format not in ["svg", "png"]:
        raise ValueError('Matplotlib format can be only "svg" or "png".')

    plot_format = merge_values(primary_value=plot_format, secondary_value=get_parameter_value('matplotlib_format'))
    embedded = merge_values(primary_value=embedded, secondary_value=get_parameter_value('matplotlib_embedded'))

    if embedded:
        if plot_format == "svg":
            f = io.BytesIO()
            fig.savefig(f, format='svg')
            img_element = f.getvalue().decode("utf-8")
        else:
            raise ValueError("Only svg format can be used for embedded matplotlib plots.")
    elif get_parameter_value('html_dir_path') and get_parameter_value('html_dir_name'):
        make_sure_dir_exists(get_parameter_value('html_dir_path'))
        img_file_name = f"img_{l_plot_index:03d}.{plot_format}"
        fig.savefig(os.path.join(get_parameter_value('html_dir_path'), img_file_name), format=plot_format,
                    bbox_inches="tight")
        img_element = f'<img src="{os.path.join(get_parameter_value("html_dir_name"), img_file_name)}">'
    else:
        raise RuntimeError("Failed to create a matplotlib image.")
    return img_element


def _prepare_altair_plot_element(fig: AltairFigType, l_plot_index: int) -> str:
    vis_id = "altairvis" + str(l_plot_index)
    img_element = f'<div id="{vis_id}"></div><script type="text/javascript">\nvar spec = {fig.to_json()};\n'
    img_element += 'var opt = {"renderer": "canvas", "actions": false};\n'
    img_element += f'vegaEmbed("#{vis_id}", spec, opt);'
    img_element += '</script>'
    return img_element


def _prepare_plotly_plot_element(fig: 'plotly.graph_objs.Figure') -> str:
    return fig.to_html(full_html=False, include_plotlyjs=False)


def _prepare_bokeh_plot_element(fig: 'bokeh.plotting.figure.Figure') -> str:
    # noinspection PyPackageRequirements
    from bokeh.embed import components  # type: ignore
    script, div = components(fig)
    return '<div>' + div + script + '</div>'


def _prepare_image_element(fig: FigType,
                           plot_index: int,
                           matplotlib_format: Optional[str] = None,
                           embedded: Optional[bool] = None,):
    # Create the html string according to the figure type.
    # (if we checked type of fig, we would have to add the libraries to requirements)
    if (type(fig).__name__ == 'Figure' and type(fig).__module__ == 'matplotlib.figure') \
            or (type(fig).__name__ == 'PairGrid' and type(fig).__module__ == 'seaborn.axisgrid'):
        img_element = _prepare_matplotlib_plot_element(fig=fig, l_plot_index=plot_index,
                                                       plot_format=matplotlib_format, embedded=embedded)
    elif type(fig).__name__ in ['Chart', 'ConcatChart', 'FacetChart', 'HConcatChart',
                                'LayerChart', 'RepeatChart', 'VConcatChart']:
        img_element = _prepare_altair_plot_element(fig=fig, l_plot_index=plot_index)
    elif type(fig).__name__ == 'Figure' and type(fig).__module__ == 'plotly.graph_objs._figure':
        img_element = _prepare_plotly_plot_element(fig=fig)
    elif type(fig).__name__ == 'Figure' and type(fig).__module__ == 'bokeh.plotting.figure':
        img_element = _prepare_bokeh_plot_element(fig=fig)
    else:
        raise ValueError(f"Unknown figure type {type(fig)}.")

    return img_element


def _plot_graph(fig: FigType,
                caption: Optional[str] = None,
                reference: Optional[Reference] = None,
                align: str = "center",
                numbered: bool = True,
                matplotlib_format: Optional[str] = None,
                embedded: Optional[bool] = None,
                hidden: bool = False) -> None:
    if not get_parameter_value('html_file_path'):
        # only when we don't print to HTML
        if type(fig).__name__ == 'Figure' and type(fig).__module__ == 'bokeh.plotting.figure':
            from bokeh.plotting import show
            show(fig)
        else:
            fig.show()
    else:
        if not hasattr(_plot_graph, 'plot_index'):
            _plot_graph.plot_index = 1
        plot_index = _plot_graph.plot_index

        anchor_link = _construct_plot_anchor_link(reference=reference, plot_ind=plot_index)
        img_element = _prepare_image_element(fig=fig, plot_index=plot_index, matplotlib_format=matplotlib_format,
                                             embedded=embedded)
        img_element += _prepare_caption_element(prefix='Figure', caption=caption, numbered=numbered,
                                                index=plot_index, anchor_link=anchor_link)
        img_html = _wrap_plot_element_by_outer_divs(img_element=img_element, align=align, hidden=hidden)

        _write_to_html(img_html)
        _plot_graph.plot_index += 1


def plot_graph(fig: FigType,
               caption: Optional[str] = None,
               reference: Optional[Reference] = None,
               align: Optional[str] = None,
               numbered: Optional[bool] = None,
               matplotlib_format: Optional[str] = None,
               embedded: Optional[bool] = None) -> None:
    """Plot a graph.

    :param fig: plot object.
    :param caption: caption of the plot.
    :param reference: reference object for link creation.
    :param align: how to align the table: "left", "center", "right".
    If None, settings from config or CLI arguments is used.
    :param numbered: should be the caption be numbered? If None, settings from config or CLI arguments is used.
    :param matplotlib_format: format in the case of matplotlib plots. Either "svg", "png", or None.
    If None, settings from config or CLI arguments is used.
    :param embedded: whether to embed the plot directly into HTML; Only for matplotlib "svg" images.
    If None, settings from config or CLI arguments is used.
    """

    align = cast(str, merge_values(primary_value=align, secondary_value=get_parameter_value('align_plots')))
    numbered = bool(merge_values(primary_value=numbered, secondary_value=get_parameter_value('numbered_plots')))

    _plot_graph(fig=fig, caption=caption, reference=reference, align=align, numbered=numbered,
                matplotlib_format=matplotlib_format, embedded=embedded, hidden=False)


def plot_multi_graph(figs: List[FigType],
                     captions: Optional[List[Optional[str]]] = None,
                     align: Optional[str] = None,
                     numbered: Optional[bool] = None) -> None:
    if captions is None:
        captions = [None] * len(figs)

    if len(figs) != len(captions):
        raise ValueError("There must be the same number of captions and figs.")

    align = cast(str, merge_values(primary_value=align, secondary_value=get_parameter_value('align_plots')))
    numbered = bool(merge_values(primary_value=numbered, secondary_value=get_parameter_value('numbered_plots')))

    if len(figs) > 0:
        if not hasattr(plot_multi_graph, 'multi_plot_index'):
            plot_multi_graph.multi_plot_index = 1
        multi_plot_index = plot_multi_graph.multi_plot_index

        b_prev_id = f"button_prev_{multi_plot_index}"
        b_next_id = f"button_next_{multi_plot_index}"
        div_id = f"image-multi-panel-{multi_plot_index}"
        disable_next = "disabled " if len(figs) == 1 else ""
        _write_to_html(f'<button id="{b_prev_id}" disabled onclick='
                       f'"previous(\'#{div_id}\', \'{b_next_id}\', \'{b_prev_id}\')">&lt;</button>')
        _write_to_html(f'<button id="{b_next_id}" {disable_next}onclick='
                       f'"next(\'#{div_id}\', \'{b_next_id}\', \'{b_prev_id}\')">&gt;</button>')
        _write_to_html(f'<div id="{div_id}">')

        for i in range(len(figs)):
            _plot_graph(fig=figs[i], caption=captions[i], align=align, numbered=numbered, hidden=i > 0)

        _write_to_html('</div>')
        plot_multi_graph.multi_plot_index += 1



import re
import sys
from pathlib import Path

from pyreball.utils.logger import get_logger

logger = get_logger()


def get_html_begin(template_path: Path, title: str, script_definitions: str, css_definitions: str) -> str:
    with open(template_path, 'r') as f:
        html_start = f.read()
        html_start = re.sub(r'{{title}}', title, html_start)
        html_start = re.sub(r'{{script_definitions}}', script_definitions, html_start)
        html_start = re.sub(r'{{css_definitions}}', css_definitions, html_start)
        return html_start


def get_html_end(template_path: Path) -> str:
    with open(template_path, 'r') as f:
        html_end = f.read()
        return html_end


def get_css(filename: str, directory: Path, page_width: int = 60) -> str:
    try:
        with open(directory / filename) as f:
            css_string = f.read()
    except OSError:
        logger.error(f'There was a problem reading file {filename} in {directory}. Try re-installing pyreball.')
        sys.exit(1)
    css_string = re.sub(r'{{page_width}}', str(int(page_width)), css_string)
    return css_string



import pytest
import logging
from pathlib import Path

from pyreball.utils.template_utils import get_css, get_html_begin, get_html_end


def test_get_html_begin(tmpdir):
    filename = "styles"
    template_path = Path(tmpdir) / filename
    with open(template_path, 'w') as f:
        f.write('<html>title: {{title}}, script: {{script_definitions}}, css: {{css_definitions}}')
    result = get_html_begin(template_path=template_path, title='t1', script_definitions='s1',
                            css_definitions='c1')
    assert result == '<html>title: t1, script: s1, css: c1'


def test_get_html_end(tmpdir):
    filename = "styles"
    template_path = Path(tmpdir) / filename
    with open(template_path, 'w') as f:
        f.write('</html>')
    result = get_html_end(template_path=template_path)
    assert result == '</html>'


def test_get_css__existing_file(tmpdir):
    filename = "styles"
    directory = Path(tmpdir)
    with open(directory / filename, 'w') as f:
        f.write('body {width: {{page_width}}%;}\n')
    styles = get_css(filename, directory, page_width=30)
    assert styles == 'body {width: 30%;}\n'


def test_get_css__non_existing_file(tmpdir, caplog):
    caplog.set_level(logging.ERROR)
    filename = "styles"
    directory = Path(tmpdir)
    with pytest.raises(SystemExit):
        get_css(filename, directory, page_width=30)
    assert 'There was a problem reading file' in caplog.text



import sys
from pathlib import Path
import argparse
import os
import re
import json
from typing import Dict, Optional, Union, Tuple, cast
import xml
from xml.dom.minidom import parseString

import pkg_resources

from pyreball.constants import (
    PATH_TO_CONFIG_LOCATION,
    DEFAULT_PATH_TO_CONFIG,
    STYLES_TEMPLATE_FILENAME,
    CONFIG_INI_FILENAME,
)
from pyreball.utils.logger import get_logger
from pyreball.utils.utils import (
    get_file_config,
    check_and_fix_parameters,
    merge_parameter_dictionaries,
    ChoiceParameter,
    IntegerParameter,
    carefully_remove_directory_if_exists,
    Substitutor,
)
from pyreball.utils.template_utils import get_css, get_html_begin, get_html_end

logger = get_logger()

# keep the indentation in the following snippets!!!
JAVASCRIPT_CHANGE_EXPAND = """
    function change_expand(button, table_id){
        var table = document.getElementById(table_id);
        if (table.classList.contains("expanded")) {
            // collapse the table
            table.style.maxHeight = "390px";
            button.innerHTML = "⟱";
        } else {
            // expand the table
            table.style.maxHeight = "none";
            button.innerHTML = "⟰";
        }
        table.classList.toggle("expanded");
    }

"""
JAVASCRIPT_ON_LOAD = """

    window.onload = function() {
    //dom not only ready, but everything is loaded
      scrollers = document.getElementsByClassName("table-scroller");

      for (i = 0; i < scrollers.length; i++) {
        if (scrollers[i].scrollHeight == scrollers[i].clientHeight) {
            // hide the expand button
            expander_id = scrollers[i].id.replace('scroller', 'expander');
            expander = document.getElementById(expander_id);
            expander.style.display = "none";
        }
      }

    };
"""

JAVASCRIPT_ROLLING_PLOTS = """

    function next(div_id, button_next_id, button_prev_id) {
        var qElems = document.querySelectorAll(div_id + '>div');
        for (var i = 0; i < qElems.length; i++) {
            if (qElems[i].style.display != 'none') {
                qElems[i].style.display = 'none';
                qElems[i + 1].style.display = 'block';
                if (i == qElems.length - 2) {
                    document.getElementById(button_next_id).disabled = true;
                }
                document.getElementById(button_prev_id).disabled = false;
                break;
            }
        }
    }

    function previous(div_id, button_next_id, button_prev_id) {
        var qElems = document.querySelectorAll(div_id + '>div');
        for (var i = 0; i < qElems.length; i++) {
            if (qElems[i].style.display != 'none') {
                qElems[i].style.display = 'none';
                qElems[i - 1].style.display = 'block';
                if (i == 1) {
                    document.getElementById(button_prev_id).disabled = true;
                }
                document.getElementById(button_next_id).disabled = false;
                break;
            }
        }
    }

"""

JAVASCRIPT_SORTABLE_TABLE = """

    $(document).ready(function () {
        $('.sortable_table').DataTable({
            "paging": false,
            "searching": false,
            "info": false,
        });
    });

"""


def replace_ids(filename: Path) -> None:
    # collect all ids in form of table-N-M
    all_table_and_img_ids = set()
    with open(filename, "r") as f:
        for line in f:
            # note that we don't need to replace only "table" ids by also "img" etc.
            results = re.findall(r'table-id[\d]+-[\d]+', line)
            if results:
                all_table_and_img_ids.update(results)
            results = re.findall(r'img-id[\d]+-[\d]+', line)
            if results:
                all_table_and_img_ids.update(results)
    replacements = []
    for element_id in all_table_and_img_ids:
        re_results = re.search(r'(.+)-(id\d+)-(\d+)', element_id)

        if re_results:
            # this must be first
            replacements.append(("ref-" + re_results.group(2), re_results.group(1) + "-" + re_results.group(3)))
            # this must be second (because it would catch the first case as well)
            replacements.append((re_results.group(2) + '(-' + re_results.group(3) + ')?', re_results.group(3)))

    # replace all table-N-M with table-M and Table N with Table M
    substitutor = Substitutor(replacements=replacements)
    modified_lines = []
    with open(filename, "r") as f:
        for line in f:
            modified_lines.append(substitutor.sub(line))

    with open(filename, "w") as f:
        f.writelines(modified_lines)


def _get_node_text(node: xml.dom.minidom.Element) -> str:
    result = []
    for child in node.childNodes:
        if child.nodeType in (xml.dom.Node.TEXT_NODE, xml.dom.Node.CDATA_SECTION_NODE):
            result.append(child.data)
        else:
            result.extend(_get_node_text(child))
    return ''.join(result)


def _parse_heading_info(line: str) -> Optional[Tuple[int, str, str]]:
    heading_pattern = r'<h(\d).+</h(\d)>'
    m = re.search(heading_pattern, line)
    if m:
        heading_level = m.group(1)

        doc = parseString(m.group(0))
        heading = doc.getElementsByTagName("h" + heading_level)[0]
        heading_id = heading.getAttribute('id')
        content = _get_node_text(heading).replace('¶', '')
        return int(heading_level), heading_id, content
    else:
        return None


def insert_heading_title_and_toc(filename: Path, include_toc: bool = True):
    # fetch all lines
    with open(filename, "r") as f:
        lines = f.readlines()

    # try to extract the title from <title> element:
    report_title = None
    for line in lines:
        m = re.match(r'^<title class="custom">([^<]*)</title>$', line)
        if m:
            report_title = m.group(1)
            break

    # get all headings in the report
    container_start_index = 0
    headings = []
    for i, line in enumerate(lines):
        if '<div class="main_container">' in line:
            container_start_index = i

        if include_toc:
            heading_info = _parse_heading_info(line)
            if heading_info:
                headings.append(heading_info)

    if len(headings) > 0 and report_title is None:
        # only when headings were collected (only when include_toc=True) and there was not title set manually
        report_title = "Table of Contents"

    lines_index = container_start_index + 1

    # prepare new HTML lines with TOC
    if report_title is not None:
        lines.insert(
            lines_index,
            f'<h1 id="toc_generated_0">{report_title}<a class="anchor-link" href="#toc_generated_0">¶</a></h1>\n'
        )
        lines_index += 1
    current_level = 1
    for h in headings:

        # do we need to add also <ul> ?
        while h[0] > current_level:
            lines.insert(lines_index, '<ul style="list-style-type:none; margin:0px">\n')
            lines_index += 1
            current_level += 1

        # do we need to add also </ul> ?
        while h[0] < current_level:
            lines.insert(lines_index, '</ul>\n')
            lines_index += 1
            current_level -= 1

        # prepare the line:
        if h[0] == 1:
            current_line = '<a href="#' + h[1] + '">' + h[2] + '</a><br/>\n'
        else:
            current_line = '<li><a href="#' + h[1] + '">' + h[2] + '</a></li>\n'
        lines.insert(lines_index, current_line)
        lines_index += 1

    # at the end, get back to level 1 if necessary
    while 1 < current_level:
        lines.insert(lines_index, '</ul>\n')
        lines_index += 1
        current_level -= 1

    with open(filename, "w") as f:
        f.writelines(lines)


parameter_specifications = [
    ChoiceParameter('--toc', choices=['yes', 'no'], default='no', help='Include table of contents.'),
    ChoiceParameter('--align-tables', choices=['left', 'center', 'right'], default='center',
                    help='Alignment of tables.'),
    ChoiceParameter('--numbered-tables', choices=['yes', 'no'], default='no', help='Number the tables.'),
    ChoiceParameter('--sortable-tables', choices=['yes', 'no'], default='no', help='Make the tables sortable.'),
    ChoiceParameter('--full-tables', choices=['yes', 'no'], default='no', help='Force all tables to be expanded.'),
    ChoiceParameter('--align-plots', choices=['left', 'center', 'right'], default='center', help='Alignment of plots.'),
    ChoiceParameter('--numbered-plots', choices=['yes', 'no'], default='no', help='Number the plots.'),
    ChoiceParameter('--matplotlib-format', choices=['png', 'svg'], default='svg', help='Format of matplotlib plots.'),
    ChoiceParameter('--matplotlib-embedded', choices=['yes', 'no'], default='no',
                    help='Whether to embedded matplotlib images directly into HTML. Only for svg format.'),
    ChoiceParameter('--numbered-headings', choices=['yes', 'no'], default='no', help='Number the headings.'),
    IntegerParameter('--page-width', boundaries=(40, 100), default=80,
                     help='Width of the page in percentage. An integer in the range 40..100.'),
    ChoiceParameter('--keep-stdout', choices=['yes', 'no'], default='no', help='Print the output to stdout too.'),
]


def parse_arguments() -> Dict[str, Optional[Union[str, int]]]:
    parser = argparse.ArgumentParser(description='Generate Python report.')
    for input_param in parameter_specifications:
        input_param.add_argument_to_parser(parser)
    parser.add_argument('--output-dir', help='Output directory. By default, the directory of the input file.')
    parser.add_argument('filename', help='Input file path.')
    args = parser.parse_args()
    return vars(args)


def get_config_directory() -> Path:
    """Get the location of the config files.

    If the configs were generated by pyreball-generate-config command, they should be found.
    If they were not generated or some of them no longer exist, the default package config will be used.
    """
    config_location_file_path = Path(PATH_TO_CONFIG_LOCATION)
    if config_location_file_path.exists():
        # the config was generated, let's find out its directory
        config_directory = Path(Path(PATH_TO_CONFIG_LOCATION).read_text())
        if not (config_directory / CONFIG_INI_FILENAME).exists() \
                or not (config_directory / STYLES_TEMPLATE_FILENAME).exists():
            logger.warning(f'{CONFIG_INI_FILENAME} or {STYLES_TEMPLATE_FILENAME} was not found in {config_directory}. '
                           f'Try re-generating the configs by pyreball-generate-config command. For now, we will '
                           f'use the default package configs.')
            config_directory = DEFAULT_PATH_TO_CONFIG
    else:
        config_directory = DEFAULT_PATH_TO_CONFIG
    return config_directory


def main() -> None:
    args_dict = parse_arguments()
    filename = Path(args_dict.pop('filename'))  # type: ignore
    output_dir = cast(Optional[str], args_dict.pop('output_dir'))
    cli_parameters = check_and_fix_parameters(parameters=args_dict, parameter_specifications=parameter_specifications,
                                              none_allowed=True)

    config_directory = get_config_directory()
    file_config_parameters = get_file_config(filename=CONFIG_INI_FILENAME, directory=config_directory,
                                             parameter_specifications=parameter_specifications)

    parameters = merge_parameter_dictionaries(primary_parameters=cli_parameters,
                                              secondary_parameters=file_config_parameters,
                                              parameter_specifications=parameter_specifications)

    if not filename.is_file():
        raise ValueError(f"File {filename} does not exist.")

    title = filename.stem
    if not output_dir:
        # use the directory of the input file
        output_dir_path = filename.resolve().parents[0]
    else:
        output_dir_path = Path(output_dir)
        output_dir_path.mkdir(parents=True, exist_ok=True)
    path_str = str(output_dir_path / title)
    os.environ["_TMP_PYREBALL_GENERATOR_PARAMETERS"] = json.dumps({**parameters, 'html_dir_path': path_str})
    html_path = Path(path_str + ".html")

    # remove the directory with images if it exists:
    carefully_remove_directory_if_exists(directory=Path(path_str))

    script_definitions = (JAVASCRIPT_CHANGE_EXPAND + JAVASCRIPT_ON_LOAD + JAVASCRIPT_SORTABLE_TABLE
                          + JAVASCRIPT_ROLLING_PLOTS)

    css_definitions = get_css(filename=STYLES_TEMPLATE_FILENAME, directory=config_directory,
                              page_width=cast(int, parameters['page_width']))

    html_begin = get_html_begin(template_path=Path(pkg_resources.resource_filename('pyreball',
                                                                                   'cfg/html_begin.template')),
                                title=title, script_definitions=script_definitions, css_definitions=css_definitions)
    html_end = get_html_end(template_path=Path(pkg_resources.resource_filename('pyreball', 'cfg/html_end.template')))

    with open(html_path, 'w') as f:
        f.write(html_begin)
    try:
        # Use {sys.executable} instead of just "python" command as it may not work correctly as a PyCharm external tool
        os.system(f"{sys.executable} {filename}")
    finally:
        with open(html_path, 'a') as f:
            f.write(html_end)

    replace_ids(html_path)
    insert_heading_title_and_toc(filename=html_path, include_toc=parameters['toc'] == 'yes')


if __name__ == '__main__':
    main()



import argparse
import os
import shutil
from pathlib import Path

from pyreball.constants import (
    PATH_TO_CONFIG_LOCATION,
    DEFAULT_PATH_TO_CONFIG,
    CONFIG_INI_FILENAME,
    STYLES_TEMPLATE_FILENAME,
)
from pyreball.utils.logger import get_logger

logger = get_logger()


def copy_config_files(output_directory: Path) -> None:
    if output_directory.exists():
        logger.warning(f"Directory {output_directory} already exists, config will be over-written.")
    output_directory.mkdir(parents=True, exist_ok=True)
    shutil.copy2(DEFAULT_PATH_TO_CONFIG / CONFIG_INI_FILENAME, output_directory)
    shutil.copy2(DEFAULT_PATH_TO_CONFIG / STYLES_TEMPLATE_FILENAME, output_directory)
    # mark down where we created the config files
    Path(PATH_TO_CONFIG_LOCATION).write_text(str(output_directory))


def main() -> None:
    parser = argparse.ArgumentParser(description='Generate Pyreball Config Files.')
    parser.add_argument('--output-dir', help='Output directory. By default, ~/.pyreball.')
    args = parser.parse_args()
    if args.output_dir:
        config_directory = Path(args.output_dir).absolute()
    else:
        config_directory = Path(os.environ["HOME"]) / '.pyreball'
    copy_config_files(output_directory=config_directory)


if __name__ == '__main__':
    main()



from setuptools import setup, find_packages

with open('requirements_test.txt') as f:
    test_required = f.read().splitlines()

with open('requirements_examples.txt') as f:
    examples_required = f.read().splitlines()

import codecs
import os.path


def read(rel_path):
    here = os.path.abspath(os.path.dirname(__file__))
    with codecs.open(os.path.join(here, rel_path), 'r') as fp:
        return fp.read()


def get_version(rel_path):
    # based on https://packaging.python.org/guides/single-sourcing-package-version/
    for line in read(rel_path).splitlines():
        if line.startswith('__version__'):
            delim = '"' if '"' in line else "'"
            return line.split(delim)[1]
    else:
        raise RuntimeError("Unable to find version string.")


setup(
    name='pyreball',
    version=get_version("pyreball/__init__.py"),
    description='pyreball',
    author='Karel Vaculik',
    author_email='vaculik.dev@gmail.com',
    packages=find_packages(exclude=('examples', 'tests')),
    package_data={'pyreball': ['cfg/html_begin.template', 'cfg/html_end.template']},
    include_package_data=True,
    tests_require=test_required + examples_required,
    entry_points={
        'console_scripts': [
            'pyreball = pyreball.__main__:main',
            'pyreball-generate-config = pyreball.config_generator:main',
        ]
    },
)
